#pragma warning(disable:4786)

#include "conv3ds.h"
#include "convtempanim.h"
#include "C3DModel.h"
#include "tangentspacecalculation.h"
#include "textureprops.h"
#include "texturecrunch.h"

bool CConv3DS::Conv3dsToIntern(const char * pFile, CTexturePropContainer * pTexPropContainer)
{
	// process file first
	C3DModel c3dsScene;
	c3dsScene.ProcessFile(pFile);

	// setup a temporary animation layout
	if(m_pTmpAnim != NULL) { delete m_pTmpAnim; m_pTmpAnim = NULL; }	
	m_pTmpAnim = new CConvTempAnim;
	m_pTmpAnim->Setup(c3dsScene);
	
	// meshes
	m_iNumMeshes = c3dsScene.GetObjNumber();
	m_pMeshes = new CInternMesh[m_iNumMeshes+1];	// +1 in case of dummy
	Vec3 * centers = new Vec3[m_iNumMeshes];

	for(int m=0; m<m_iNumMeshes; m++)
		centers[m] = ConvMeshToIntern(&m_pMeshes[m], c3dsScene.GetObject(m));

	// meshinstances
	MakeAllMeshInstances(c3dsScene, centers);
	delete [] centers;
	
	// lights
	MakeAllCameras(c3dsScene);

	// cameras
	MakeAllLights(c3dsScene);
	

	// materials
	m_iNumMats = DetermineNumMaterials(c3dsScene);
	m_pMaterials = new CInternMaterial[m_iNumMats];
	DistributeMatNames(c3dsScene);
	DistributeMatToSurf(c3dsScene);
	DistributeTextureSequences(c3dsScene, pTexPropContainer, pFile);
	
	// Create dummies
	const bool bFoundDummy = c3dsScene.GetTrackInfo(string("$$$DUMMY"))!=NULL;
	if(bFoundDummy)
	{
		m_pMeshes[m_iNumMeshes].m_iNumSurfs = 0;
		m_pMeshes[m_iNumMeshes].m_pSurfaces = NULL;
		m_pMeshes[m_iNumMeshes].m_sMeshName = new char[strlen("$$$DUMMY")+1];
		strcpy(m_pMeshes[m_iNumMeshes++].m_sMeshName, "$$$DUMMY");
	}

	MakeAllDummyInstances(c3dsScene);


	// Fetch Scene Name
	FetchNameOfScene(pFile);

	// set anim time
	m_fAnimTime = (c3dsScene.GetFrameEnd()-c3dsScene.GetFrameBegin()+1) / m_fFPS;
	

	return true;
}

const Vec3 CConv3DS::ConvMeshToIntern(CInternMesh * pIntMesh, C3DObject * p3dsObj)
{
	// setup transformation
	const Vec3 v1(p3dsObj->transform[0][0], p3dsObj->transform[0][1], p3dsObj->transform[0][2]);
	const Vec3 v2(p3dsObj->transform[1][0], p3dsObj->transform[1][1], p3dsObj->transform[1][2]);
	const Vec3 v3(p3dsObj->transform[2][0], p3dsObj->transform[2][1], p3dsObj->transform[2][2]);
	const Vec3 point(p3dsObj->translate[0], p3dsObj->translate[1], p3dsObj->translate[2]);

	// Convert transformation to intermediate format.
	Mat33 m33;
	SetRow(&m33, 0, v1);
	SetRow(&m33, 1, v2);
	SetRow(&m33, 2, v3);

	// perform the most fucked up hack
	// in the world due to kinetics insanity
	Mat44 m44;
	SetColumn(&m44, 0, GetColumn(m33, 0)); 
	SetColumn(&m44, 1, GetColumn(m33, 1));
	SetColumn(&m44, 2, GetColumn(m33, 2));
	SetColumn(&m44, 3, point);
	SetRow(&m44, 3, Vec4(0,0,0,1));
	m44 = ~m44;	// invert
	for(int k=0; k<p3dsObj->vertices.size(); k++)
	{
		const float x = p3dsObj->vertices[k].x;
		const float y = p3dsObj->vertices[k].y;
		const float z = p3dsObj->vertices[k].z;
		const Vec3 v = m44 * Vec4(x, y, z, 1);
		p3dsObj->vertices[k].x = v.x;
		p3dsObj->vertices[k].y = v.y;
		p3dsObj->vertices[k].z = v.z;
	}


	// not pretty, I'm going to hack into the 3dsObj class
	// to make sure various data isn't missing
	assert( p3dsObj->descriptors.size() <= p3dsObj->faces.size() );
	bool * pTrngBool = NULL;
	pTrngBool = new bool[p3dsObj->faces.size()];
	for(int t=0; t<p3dsObj->faces.size(); t++)
		pTrngBool[t]=false;
	for(int d=0; d<p3dsObj->descriptors.size(); d++)
		pTrngBool[p3dsObj->descriptors[d].m_Number]=true;
	for(t=0; t<p3dsObj->faces.size(); t++)
		if(pTrngBool[t]==false)
			p3dsObj->descriptors.push_back(CDescriptor(string("dummy_material"), t));
	delete [] pTrngBool;
	if(p3dsObj->texturePoints.size() == 0)	// and texture coordinates
		for(int i=0; i<p3dsObj->vertices.size(); i++)
			p3dsObj->texturePoints.push_back(VECTOR(0, 0, 1));
	while(p3dsObj->smoothrefs.size() < p3dsObj->faces.size())
		p3dsObj->smoothrefs.push_back(0);
			

	// allocate room for surfaces
	int numSurfs = DetermineNumSurfaces(p3dsObj);
	pIntMesh->m_pSurfaces = new CInternSurface[numSurfs];
	pIntMesh->m_iNumSurfs = numSurfs;
	pIntMesh->m_sMeshName = new char[strlen(p3dsObj->m_sObjName.c_str())+1];
	strcpy(pIntMesh->m_sMeshName, p3dsObj->m_sObjName.c_str());
	
	// setup facelist and the geometry
	int32 ** descs;	// pick up a descriptor for every face added to each surface
	const int32 max_trng = CreateFaceLists(pIntMesh, p3dsObj, &descs);

	Vec3 * pNorms = new Vec3[p3dsObj->faces.size()*3];	// max len
	SInternFace * trngNrms = new SInternFace[p3dsObj->faces.size()];
	CreateNormals(p3dsObj, pNorms, trngNrms);
	Vec3 * pTang = new Vec3[p3dsObj->faces.size()*3];	// max len
	Vec3 * pBino = new Vec3[p3dsObj->faces.size()*3];	// max len
	SInternFace * trngTang = new SInternFace[p3dsObj->faces.size()];
	CreateTangBinVectors(p3dsObj, pNorms, trngNrms, pTang, pBino, trngTang );
	CreateGeometryData(pIntMesh, p3dsObj, pNorms, trngNrms, pTang, pBino, trngTang, (const int32 **) descs, max_trng);
	delete [] pNorms; delete [] trngNrms;
	delete [] pTang; delete [] pBino; delete [] trngTang;
	for(int32 s=0; s<pIntMesh->m_iNumSurfs; s++) delete [] descs[s];
	delete [] descs;

	// calculate bounding box
	Vec3 vCen(0,0,0);
	if(pIntMesh->GetNumSurfaces()>0)
	{
		Vec3 vMin = pIntMesh->GetSurface(0).GetVertex(0), vMax = vMin;
		for(int s=0; s<pIntMesh->GetNumSurfaces(); s++)
			for(int v=0; v<pIntMesh->GetSurface(s).GetArrayLen(); v++)
			{
				const Vec3 pnt = pIntMesh->GetSurface(s).GetVertex(v);

				if(vMin.x > pnt.x) vMin.x = pnt.x;
				else if(vMax.x < pnt.x) vMax.x = pnt.x;

				if(vMin.y > pnt.y) vMin.y = pnt.y;
				else if(vMax.y < pnt.y) vMax.y = pnt.y;

				if(vMin.z > pnt.z) vMin.z = pnt.z;
				else if(vMax.z < pnt.z) vMax.z = pnt.z;
			}
		vCen = (vMin+vMax)*0.5f;

		// centralize mesh
		for(s=0; s<pIntMesh->GetNumSurfaces(); s++)
			for(int v=0; v<pIntMesh->GetSurface(s).GetArrayLen(); v++)
				pIntMesh->GetSurface(s).m_pVerts[v] -= vCen;
	}

	// convert systems
	const Mat33 mAxisUpMat = GetAxisUpMatrix();
	for(s=0; s<pIntMesh->GetNumSurfaces(); s++)
		for(int v=0; v<pIntMesh->GetSurface(s).GetArrayLen(); v++)
		{
			pIntMesh->GetSurface(s).m_pVerts[v] = mAxisUpMat * pIntMesh->GetSurface(s).m_pVerts[v];
			pIntMesh->GetSurface(s).m_pNormals[v] = mAxisUpMat * pIntMesh->GetSurface(s).m_pNormals[v];
			pIntMesh->GetSurface(s).m_pTangents[v] = mAxisUpMat * pIntMesh->GetSurface(s).m_pTangents[v];
			pIntMesh->GetSurface(s).m_pBiNorms[v] = mAxisUpMat * pIntMesh->GetSurface(s).m_pBiNorms[v];
		}
		
	return vCen;
}

void CConv3DS::MakeInternMeshInstance(CInternMeshInstance * pMeshInst, CInternMesh * pMesh, const string name, C3DObject * p3dsObj, const Vec3 vCen, bool bFoundAnim)
{
	// setup transformation
	const Vec3 v1(p3dsObj->transform[0][0], p3dsObj->transform[0][1], p3dsObj->transform[0][2]);
	const Vec3 v2(p3dsObj->transform[1][0], p3dsObj->transform[1][1], p3dsObj->transform[1][2]);
	const Vec3 v3(p3dsObj->transform[2][0], p3dsObj->transform[2][1], p3dsObj->transform[2][2]);
	const Vec3 point(p3dsObj->translate[0], p3dsObj->translate[1], p3dsObj->translate[2]);

	// Convert transformation to intermediate format.
	Mat33 m33;
	SetRow(&m33, 0, v1);
	SetRow(&m33, 1, v2);
	SetRow(&m33, 2, v3);

	// set mesh
	//char digits[256];	// seem reasonable to assume that 256 would be an upperbound
	//sprintf(digits, "%d", num_insts);
	//pMeshInst->m_sMeshInstName = *new string( pMesh->GetName() + string("Inst_") + string(digits) );
	pMeshInst->m_sMeshInstName = new char[strlen(name.c_str())+1];//*new string( name.c_str() );
	strcpy(pMeshInst->m_sMeshInstName, name.c_str());
	pMeshInst->m_pMesh = pMesh;
	
	// Look for spline, if we don't find one then make a default one
	CreateTransform(pMeshInst, vCen, m33, point, bFoundAnim);
}

void CConv3DS::MakeAllMeshInstances(C3DModel &c3dsScene, Vec3 * centers)
{
	// anims pointing to an existing mesh count as an instance
	// meshes with no existing anim also count as an instance
	int32 max_insts = m_iNumMeshes + m_pTmpAnim->GetNumMeshAnims(), loc_mesh;
	m_iNumMeshInstances = 0;
	m_pMeshInstances = new CInternMeshInstance[max_insts];	
	for(int m=0; m<m_pTmpAnim->GetNumMeshAnims(); m++)
		//	anim_has_mesh()
		if(m_pTmpAnim->SetMesh(m) && 
			(loc_mesh=GetMeshIndex(c3dsScene, m_pTmpAnim->GetName())) != -1)
		{
			CInternMesh * pMesh = &m_pMeshes[loc_mesh];
			C3DObject * p3dsObj = c3dsScene.GetObject(loc_mesh);
			const Vec3 vCen = centers[loc_mesh];
			const string name = m_pTmpAnim->GetTrackInstName().empty()?m_pTmpAnim->GetName():m_pTmpAnim->GetTrackInstName();
			MakeInternMeshInstance(&m_pMeshInstances[m_iNumMeshInstances], pMesh, name, p3dsObj, vCen, true);
			++m_iNumMeshInstances;
		}
	for(m=0; m<m_iNumMeshes; m++)
	{
		C3DObject * p3dsObj = c3dsScene.GetObject(m);
		// mesh_has_no_anim()
		if(!m_pTmpAnim->SetMesh( p3dsObj->m_sObjName ) )
		{
			CInternMesh * pMesh = &m_pMeshes[m];
			const Vec3 vCen = centers[m];
			const string name = p3dsObj->m_sObjName;
			MakeInternMeshInstance(&m_pMeshInstances[m_iNumMeshInstances], pMesh, name, p3dsObj, vCen, false);
			++m_iNumMeshInstances;
		}
	}
	assert(m_iNumMeshInstances <= max_insts);
}

void CConv3DS::MakeAllDummyInstances(C3DModel &c3dsScene)
{
	// dummy anims always point to the dummy mesh at the end
	for(int m=0; m<m_pTmpAnim->GetNumMeshAnims(); m++)
		//	anim_has_mesh()
		if(m_pTmpAnim->SetMesh(m) && m_pTmpAnim->GetName()==string("$$$DUMMY") )
		{
			CInternMeshInstance * pMeshInst = &m_pMeshInstances[m_iNumMeshInstances];
			const string name = m_pTmpAnim->GetTrackInstName().empty()?m_pTmpAnim->GetName():m_pTmpAnim->GetTrackInstName();
			pMeshInst->m_sMeshInstName = new char[strlen(name.c_str())+1];//*new string( name.c_str() );
			strcpy(pMeshInst->m_sMeshInstName, name.c_str());
			pMeshInst->m_pMesh = &m_pMeshes[m_iNumMeshes-1];
			
			// Look for spline, if we don't find one then make a default one
			Mat33 rot; LoadIdentity(&rot);
			CreateTransform(pMeshInst, Vec3(0,0,0), rot, Vec3(0,0,0), true);
			++m_iNumMeshInstances;
		}
	const int32 max_insts = m_iNumMeshes + m_pTmpAnim->GetNumMeshAnims();
	assert(m_iNumMeshInstances <= max_insts);
}


int CConv3DS::DetermineNumSurfaces(C3DObject * p3dsObj)
{
	assert( p3dsObj->descriptors.size() == p3dsObj->faces.size() );
	int numSurfs=0;
	
	for(int d=0; d<(int) p3dsObj->descriptors.size(); d++)
	{
		// look for a previous copy
		int32 d2 = 0;
		bool bFoundCopy = false;
		while(d2<d && !bFoundCopy)
		{
			bFoundCopy = p3dsObj->descriptors[d].m_Name==p3dsObj->descriptors[d2].m_Name;
			++d2;
		}

		// if face is valid and name was unique
		if(!bFoundCopy && isFaceValid(p3dsObj->faces[p3dsObj->descriptors[d].m_Number], p3dsObj))
			++numSurfs;
	}

	return numSurfs;
}

string CConv3DS::FetchSurfName(int32 iSurfNum, C3DObject * p3dsObj)
{
	assert( p3dsObj->descriptors.size() == p3dsObj->faces.size() );
	int numSurfs=0, d=0;
	bool bFoundSurface = false;
	
	while(d<(int) p3dsObj->descriptors.size() && !bFoundSurface)
	{
		// look for a previous copy
		int32 d2 = 0;
		bool bFoundCopy = false;
		while(d2<d && !bFoundCopy)
		{
			bFoundCopy = p3dsObj->descriptors[d].m_Name==p3dsObj->descriptors[d2].m_Name;
			++d2;
		}

		// if face is valid and name was unique
		if(!bFoundCopy && isFaceValid(p3dsObj->faces[p3dsObj->descriptors[d].m_Number], p3dsObj))
			++numSurfs;

		// if we found a match
		if(numSurfs == iSurfNum+1)
			bFoundSurface = true;
		else
			++d;	// increment
	}

	assert(bFoundSurface);

	return p3dsObj->descriptors[d].m_Name;
}

bool CConv3DS::isFaceValid(C3DFace &face, C3DObject * p3dsObj)
{
	Vec3 p1(p3dsObj->vertices[face.A].x, p3dsObj->vertices[face.A].y, p3dsObj->vertices[face.A].z);
	Vec3 p2(p3dsObj->vertices[face.B].x, p3dsObj->vertices[face.B].y, p3dsObj->vertices[face.B].z);
	Vec3 p3(p3dsObj->vertices[face.C].x, p3dsObj->vertices[face.C].y, p3dsObj->vertices[face.C].z);
		
	return p1!=p2 && p1!=p3 && p2!=p3;
}

int32 CConv3DS::DetermineNumMaterials(C3DModel &c3dmod)
{
	int32 numMats = 0;
	for(int m=0; m<m_iNumMeshes; m++)
	{
		C3DObject * p3dsObj = c3dmod.GetObject(m);
		
		for(int d=0; d<(int) p3dsObj->descriptors.size(); d++)
		{
			// look for a previous copy
			int32 m2 = 0;
			bool bFoundCopy = false;
			while(m2<=m)
			{
				int32 d2 = 0;
				C3DObject * p3dsObj2 = c3dmod.GetObject(m2);
				const int32 lim = (m2==m) ? d : (int32) p3dsObj2->descriptors.size();
				while(d2<lim && !bFoundCopy)
				{
					bFoundCopy = p3dsObj->descriptors[d].m_Name==p3dsObj2->descriptors[d2].m_Name;
					++d2;
				}
				++m2;
			}

			// if face is valid and name was unique
			if(!bFoundCopy && isFaceValid(p3dsObj->faces[p3dsObj->descriptors[d].m_Number], p3dsObj))
				++numMats;
		}
	}
		
	return numMats;
}

void CConv3DS::DistributeMatNames(C3DModel &c3dmod)
{
	int32 numMats = 0;
	for(int m=0; m<m_iNumMeshes; m++)
	{
		C3DObject * p3dsObj = c3dmod.GetObject(m);
			
		for(int d=0; d<(int) p3dsObj->descriptors.size(); d++)
		{
			// look for a previous copy
			int32 m2 = 0;
			bool bFoundCopy = false;
			while(m2<=m)
			{
				int32 d2 = 0;
				C3DObject * p3dsObj2 = c3dmod.GetObject(m2);
				const int32 lim = (m2==m) ? d : (int32) p3dsObj2->descriptors.size();
				while(d2<lim && !bFoundCopy)
				{
					bFoundCopy = p3dsObj->descriptors[d].m_Name==p3dsObj2->descriptors[d2].m_Name;
					++d2;
				}
				++m2;
			}

			// if face is valid and name was unique
			if(!bFoundCopy && isFaceValid(p3dsObj->faces[p3dsObj->descriptors[d].m_Number], p3dsObj))
			{
				m_pMaterials[numMats].m_matName = new char[strlen(p3dsObj->descriptors[d].m_Name.c_str())+1];//*new string( p3dsObj->descriptors[d].m_Name.c_str() );
				strcpy(m_pMaterials[numMats].m_matName, p3dsObj->descriptors[d].m_Name.c_str());
				++numMats;
			}
		}
	}

	assert(numMats == m_iNumMats);
}

void CConv3DS::DistributeMatToSurf(C3DModel &c3dmod)
{
	for(int32 m=0; m<m_iNumMeshes; m++)
	{
		CInternMesh &mesh = m_pMeshes[m];
		for(int32 s=0; s<mesh.m_iNumSurfs; s++)
		{
			CInternSurface &surf = mesh.m_pSurfaces[s];
			string surf_name = FetchSurfName(s, c3dmod.GetObject(m));

			int32 mat=0;
			bool bMatched = false;
			while(mat<m_iNumMats && !bMatched)
				bMatched = strcmp(surf_name.c_str(), m_pMaterials[mat++].m_matName)==0;
			assert(bMatched);
			surf.m_pMaterial = &m_pMaterials[mat-1];
		}
	}
}

void CConv3DS::DistributeTextureSequences(const C3DModel &c3dsScene, CTexturePropContainer * pTexPropContainer, const char input_file[])
{
	// add path of input 3ds file to texturepaths
	// Disconnect filename from path
	const int32 str_len = strlen(input_file);
	int32 end_path = str_len - 1;
	while(end_path>=0 && !( (input_file[end_path]=='\\') || (input_file[end_path]=='/') || (input_file[end_path]==':') ))
		--end_path;
	if(end_path >=0)	// -1 means we never ran into a path
		pTexPropContainer->AddTextureFolder(input_file, end_path+1);	// end_path+1 is length of path

	// lookup
	CTextureCrunch tex_cruncher( *pTexPropContainer );
	for(int m=0; m<m_iNumMats; m++)
	{
		// locate this material in .3ds scene
		const char * mat_name = m_pMaterials[m].GetName();
		bool bFound = false;
		int n=0;
		while(n<c3dsScene.GetMatNumber() && !bFound)
			if(strcmp(mat_name, c3dsScene.GetMaterial(n).m_sMatName.c_str())==0)
				bFound = true;
			else
				++n;

		assert(bFound || strcmp(mat_name, "dummy_material")==0);

		// Fetch sequences
		if(bFound)
		{
			m_pMaterials[m].m_pDiffTextureSeq = tex_cruncher.FetchTextureSeq( c3dsScene.GetMaterial(n).m_sDiffMapName.c_str(), m_fFPS );
			m_pMaterials[m].m_pSpecTextureSeq = tex_cruncher.FetchTextureSeq( c3dsScene.GetMaterial(n).m_sSpecMapName.c_str(), m_fFPS );
			m_pMaterials[m].m_pReflTextureSeq = tex_cruncher.FetchTextureSeq( c3dsScene.GetMaterial(n).m_sReflMapName.c_str(), m_fFPS );
			m_pMaterials[m].m_pBumpTextureSeq = tex_cruncher.FetchTextureSeq( c3dsScene.GetMaterial(n).m_sBumpMapName.c_str(), m_fFPS );
		}
	}

	// remove tex_folder
	if(end_path >=0)	// -1 means we never ran into a path
		pTexPropContainer->RemoveLastTexFolder();
}

void CConv3DS::FetchNameOfScene(const char * pFile)
{
	const int32 str_len = strlen(pFile);
	const int32 iDot3DSlen = 4;	// ".3ds"
	assert(	str_len>iDot3DSlen &&	
			( (strcmp(pFile+(str_len-iDot3DSlen), ".3ds")==0) || 
			  (strcmp(pFile+(str_len-iDot3DSlen), ".3DS")==0) ) );

	// Determine length of name
	int32 loc = str_len - (iDot3DSlen+1);
	while(loc>=0 && !( (pFile[loc]=='\\') || (pFile[loc]=='/') || (pFile[loc]==':') ))
		--loc;
	const int32 new_len = (str_len - iDot3DSlen) - (loc+1);	// loc can be -1 (if file is in current directory)
			
	// fetch name
	m_sSceneName = NULL;
	m_sSceneName = new char[new_len+1];
	strncpy(m_sSceneName, &pFile[loc+1], new_len);
	m_sSceneName[new_len]='\0';			// terminate
}


////////////////////////////////////////////////////////////////////
//////////////////////// geometry creation /////////////////////////

int32 CConv3DS::CreateFaceLists(CInternMesh * pIntMesh, C3DObject * p3dsObj, int32 *** pdescs)
{
	// create surfaces
	const int32 numSurfs = pIntMesh->m_iNumSurfs;
	*pdescs = new int32 *[numSurfs];
	int32 max_trng = 0;
	for(int s=0; s<numSurfs; s++)
	{
		CInternSurface &surf = pIntMesh->m_pSurfaces[s];
		const string surf_name = FetchSurfName(s, p3dsObj);

		// detect amount of triangles on current surface
		surf.m_iNumFaces = 0;
		for(int f=0; f<(int) p3dsObj->descriptors.size(); f++)
		{
			const CDescriptor &desc = p3dsObj->descriptors[f];
			if(desc.m_Name==surf_name && isFaceValid(p3dsObj->faces[desc.m_Number], p3dsObj))
				++surf.m_iNumFaces;	
		}
		if(surf.m_iNumFaces>max_trng) max_trng = surf.m_iNumFaces;

		// allocate space for triangles
		surf.m_pFaces = new SInternFace[surf.m_iNumFaces];
		(*pdescs)[s] = new int32[surf.m_iNumFaces];

		// read faces
		int32 face = 0;
		for(f=0; f<(int) p3dsObj->descriptors.size(); f++)
		{
			const CDescriptor &desc = p3dsObj->descriptors[f];
			if(desc.m_Name==surf_name && isFaceValid(p3dsObj->faces[desc.m_Number], p3dsObj))
			{
				((*pdescs)[s])[face] = (int32) desc.m_Number;
				
				// incement
				++face;
			}
		}

	}

	return max_trng;
}

/*void CConv3DS::CreateNormals(C3DObject * p3dsObj, Vec3 * pFinalNormals, SInternFace * trngNrms)
{
	// build the Normals
	Vec3 * pTmpVerts = new Vec3[p3dsObj->faces.size()*3];	// max len
	unsigned int * tmpSmoothRefs = new unsigned int[p3dsObj->faces.size()*3];
	int32 uniquenormals = 0;

	// setup a temporary index list
	for(int f=0; f<(int) p3dsObj->faces.size(); f++)
	{
		const unsigned int smoothRef = p3dsObj->smoothrefs[f];
		int32 indices[] = {p3dsObj->faces[f].A, p3dsObj->faces[f].B, p3dsObj->faces[f].C};
		
		for(int i=0; i<3; i++)
		{
			const Vec3 vP(p3dsObj->vertices[indices[i]].x, p3dsObj->vertices[indices[i]].y, p3dsObj->vertices[indices[i]].z);
			
			// look for a duplicate
			bool bFoundCpy = false;
			int k=0;
			if(smoothRef!=0)
				while(k<uniquenormals && !bFoundCpy)
				{
					if(pTmpVerts[k]==vP && tmpSmoothRefs[k]==smoothRef)
						bFoundCpy = true;
					else
						++k;	// increment
				}

			// if unique 
			if(!bFoundCpy)
			{
				k = uniquenormals;
				++uniquenormals;
			}

			// deliver data to tmp buffer with a 
			// reordered index for the triangle indices
			pTmpVerts[k] = vP; tmpSmoothRefs[k] = smoothRef; indices[i] = k;
		}

		trngNrms[f] = SInternFace(indices[0], indices[1], indices[2]);
	}

	// fill up list of unique normals
	for(int n=0; n<uniquenormals; n++) pFinalNormals[n]=Vec3(0,0,0);
	for(int f=0; f<(int) p3dsObj->faces.size(); f++)
	{
		if(isFaceValid(p3dsObj->faces[f], p3dsObj))
		{
			const int i1 = trngNrms[f].i1;
			const int i2 = trngNrms[f].i2;
			const int i3 = trngNrms[f].i3;
			const Vec3 pnts[]={pTmpVerts[i1], pTmpVerts[i2], pTmpVerts[i3]};
			const unsigned int grpsmoothref = p3dsObj->smoothrefs[f];
			const Vec3 vN = Cross(pnts[1]-pnts[0], pnts[2]-pnts[0]);
			const int indices[] = {i1, i2, i3};
			
			for(int v=0; v<uniquenormals; v++)
				for(int i=0; i<3; i++)
					if(	(pTmpVerts[v]==pnts[i] && (grpsmoothref&tmpSmoothRefs[v])!=0) || v==indices[i] )
						pFinalNormals[v]+= vN;
		}
	}
	delete [] pTmpVerts; delete [] tmpSmoothRefs;
	for(int n=0; n<uniquenormals; n++)
		if(pFinalNormals[n]!=Vec3(0,0,0))
			pFinalNormals[n] = Normalize(pFinalNormals[n]);
}*/

void CConv3DS::CreateNormals(C3DObject * p3dsObj, Vec3 * pFinalNormals, SInternFace * trngNrms)
{
	// build the Normals
	Vec3 * pTmpVerts = new Vec3[p3dsObj->faces.size()*3];	// max len
	unsigned int * tmpSmoothRefs = new unsigned int[p3dsObj->faces.size()*3];
	memset(tmpSmoothRefs, 0, (p3dsObj->faces.size()*3) * sizeof(unsigned int) );
	int32 uniquenormals = 0;

	// setup a temporary index list
	for(int f=0; f<(int) p3dsObj->faces.size(); f++)
	{
		const unsigned int smoothRef = p3dsObj->smoothrefs[f];
		int32 indices[] = {p3dsObj->faces[f].A, p3dsObj->faces[f].B, p3dsObj->faces[f].C};
		
		for(int i=0; i<3; i++)
		{
			const Vec3 vP(p3dsObj->vertices[indices[i]].x, p3dsObj->vertices[indices[i]].y, p3dsObj->vertices[indices[i]].z);
			
			// look for a duplicate
			bool bFoundCpy = false;
			int k=0;
			while(k<uniquenormals && !bFoundCpy)
			{
				if(pTmpVerts[k]==vP && (tmpSmoothRefs[k]&smoothRef)!=0)
					bFoundCpy = true;
				else
					++k;	// increment
			}

			// if unique 
			if(!bFoundCpy)
			{
				assert(k == uniquenormals);
				++uniquenormals;
				pTmpVerts[k] = vP; 
			}
			else	// found partner
			{
				assert(pTmpVerts[k] == vP );
				for(int j=k+1; j<uniquenormals; j++)	// look for others to merge with
				{
					assert((tmpSmoothRefs[j]&tmpSmoothRefs[k])==0 || pTmpVerts[j]!=vP);
					if(pTmpVerts[j]==vP && (tmpSmoothRefs[j]&smoothRef)!=0)
					{
						// first go through all triangles before f,
						// that point to j and make them point to k instead,
						// then overwrite j with the last entry of the arrays
						for(int f2=0; f2<=f; f2++)
						{
							const int32 num_verts = (f2<f)?3:i;
							int32 * pIndices[] = {&trngNrms[f2].i1, &trngNrms[f2].i2, &trngNrms[f2].i3};
							if(f2==f) { pIndices[0] = &indices[0]; pIndices[1] = &indices[1]; pIndices[2] = &indices[2]; }
							for(int i2=0; i2<num_verts; i2++)
								if(*pIndices[i2]==j) *pIndices[i2] = k;
								else if(*pIndices[i2]==(uniquenormals-1)) *pIndices[i2] = j;
						}

						// then add smoothing group contribution
						tmpSmoothRefs[k] |= tmpSmoothRefs[j];

						// then overwrite entry j in tmpSmoothRefs and pTmpVerts
						assert(uniquenormals>0);
						tmpSmoothRefs[j] = tmpSmoothRefs[uniquenormals-1];
						pTmpVerts[j] = pTmpVerts[uniquenormals-1];
						--uniquenormals;

						// we've deleted an entry, need to rewind j
						--j;
					}
				}
			}

			// deliver data to tmp buffer with a 
			// reordered index for the triangle indices
			tmpSmoothRefs[k] |= smoothRef; indices[i] = k;
		}

		trngNrms[f] = SInternFace(indices[0], indices[1], indices[2]);
	}

	// fill up list of unique normals
	for(int n=0; n<uniquenormals; n++) pFinalNormals[n]=Vec3(0,0,0);

	for(f=0; f<(int) p3dsObj->faces.size(); f++)
	{
		if(isFaceValid(p3dsObj->faces[f], p3dsObj))
		{
			const int i1 = trngNrms[f].i1;
			const int i2 = trngNrms[f].i2;
			const int i3 = trngNrms[f].i3;
			const Vec3 pnts[]={pTmpVerts[i1], pTmpVerts[i2], pTmpVerts[i3]};

			Vec3 e1 = pnts[1]-pnts[0];
			Vec3 e2 = pnts[2]-pnts[1];
			Vec3 e3 = pnts[0]-pnts[2];
			Vec3 vN = Cross(e3, e1);
			if(e1!=Vec3(0,0,0)) e1 = Normalize(e1);
			if(e2!=Vec3(0,0,0)) e2 = Normalize(e2);
			if(e3!=Vec3(0,0,0)) e3 = Normalize(e3);
			if(vN!=Vec3(0,0,0)) vN = Normalize(vN);

			const float d1 = e1*(-e3);
			const float d2 = e2*(-e1);
			const float d3 = e3*(-e2);

			const float a1 = acosf( (d1<-1)?-1:((d1>1)?1:d1) );
			const float a2 = acosf( (d2<-1)?-1:((d2>1)?1:d2) );
			const float a3 = acosf( (d3<-1)?-1:((d3>1)?1:d3) );

			pFinalNormals[i1] += (a1*vN);
			pFinalNormals[i2] += (a2*vN);
			pFinalNormals[i3] += (a3*vN);
		}
	}
	delete [] pTmpVerts; delete [] tmpSmoothRefs;
	for(n=0; n<uniquenormals; n++)
		if(pFinalNormals[n]!=Vec3(0,0,0))
			pFinalNormals[n] = Normalize(pFinalNormals[n]);
}

void CConv3DS::CreateTangBinVectors(C3DObject * p3dsObj, const Vec3 vNormals[], const SInternFace trngNorms[], Vec3 * vS, Vec3 * vT, SInternFace * trngTang)
{	
	// Create a class containing data
	CTriangleInputProxy myProxy;

	
	// Setup indexlists of unique components
	myProxy.m_pVerts = NULL; myProxy.m_pVerts=new float[3*p3dsObj->vertices.size()];
	myProxy.m_pSTs = NULL; myProxy.m_pSTs=new float[2*p3dsObj->texturePoints.size()];
	Vec3 * pNormals = NULL; pNormals=new Vec3[3*p3dsObj->faces.size()];
	myProxy.m_pPosIndices = NULL; myProxy.m_pPosIndices=new int32[3*p3dsObj->faces.size()];
	myProxy.m_pSTIndices = NULL; myProxy.m_pSTIndices=new int32[3*p3dsObj->faces.size()];
	myProxy.m_pNormIndices = NULL; myProxy.m_pNormIndices=new int32[3*p3dsObj->faces.size()];
	assert(myProxy.m_pVerts != NULL);
	assert(myProxy.m_pSTs != NULL);
	assert(pNormals != NULL);
	assert(myProxy.m_pPosIndices != NULL);
	assert(myProxy.m_pSTIndices != NULL);
	assert(myProxy.m_pNormIndices != NULL);

	int32 iNumUniqueVerts = 0;
	int32 iNumUniqueSTs = 0;
	int32 iNumUniqueNorms = 0;
	int32 iNumFaces = 0;
	for(int f=0; f<p3dsObj->faces.size(); f++)
	{
		if(isFaceValid(p3dsObj->faces[f], p3dsObj))
		{
			int32 indices_vert[] = {p3dsObj->faces[f].A, p3dsObj->faces[f].B, p3dsObj->faces[f].C};
			int32 indices_st[3];
			int32 indices_norm[] = {trngNorms[f].i1, trngNorms[f].i2, trngNorms[f].i3};

			// fill up index list for Verts and STs
			for(int i=0; i<3; i++)
			{
				int32 index = indices_vert[i];
				Vec3 p(p3dsObj->vertices[index].x, p3dsObj->vertices[index].y, p3dsObj->vertices[index].z);
				Vec3 stq(p3dsObj->texturePoints[index].x, p3dsObj->texturePoints[index].y, 1);
				Vec3 n = vNormals[indices_norm[i]];
				
				// look for a duplicate vertex
				bool bFoundCpy = false;
				int k=0;
				while(k<iNumUniqueVerts && !bFoundCpy)
				{
					if(Vec3(myProxy.m_pVerts[k*3+0], myProxy.m_pVerts[k*3+1], myProxy.m_pVerts[k*3+2])==p)
						bFoundCpy = true;
					else
						++k;	// increment
				}

				// if unique 
				if(!bFoundCpy)
				{
					k = iNumUniqueVerts;
					++iNumUniqueVerts;

					myProxy.m_pVerts[k*3+0]=p.x;
					myProxy.m_pVerts[k*3+1]=p.y;
					myProxy.m_pVerts[k*3+2]=p.z;
				}

				indices_vert[i]=k;

				// look for a duplicate stq
				bFoundCpy = false;
				k=0;
				while(k<iNumUniqueSTs && !bFoundCpy)
				{
					if(Vec3(myProxy.m_pSTs[k*2+0], myProxy.m_pSTs[k*2+1], 1)==stq)
						bFoundCpy = true;
					else
						++k;	// increment
				}

				// if unique 
				if(!bFoundCpy)
				{
					k = iNumUniqueSTs;
					++iNumUniqueSTs;

					myProxy.m_pSTs[k*2+0]=stq.x;
					myProxy.m_pSTs[k*2+1]=stq.y;
				}
				
				indices_st[i]=k;

				// look for a duplicate vertex normal
				bFoundCpy = false;
				k=0;
				while(k<iNumUniqueNorms && !bFoundCpy)
				{
					if(pNormals[k]==n)
						bFoundCpy = true;
					else
						++k;	// increment
				}

				// if unique 
				if(!bFoundCpy)
				{
					k = iNumUniqueNorms;
					++iNumUniqueNorms;
					
					pNormals[k]=n;
				}

				indices_norm[i]=k;
			}

			// fillup ST, vert and norm indices
			for(int q=0; q<3; q++)
			{
				myProxy.m_pPosIndices[iNumFaces*3+q]=indices_vert[q];
				myProxy.m_pSTIndices[iNumFaces*3+q]=indices_st[q];
				myProxy.m_pNormIndices[iNumFaces*3+q]=indices_norm[q];
			}

			// proceed
			++iNumFaces;
		}
	}
	myProxy.m_iNumTriangles = iNumFaces;
	
	// Calculate the tangent vectors and binormals
 	CTangentSpaceCalculation<CTriangleInputProxy> MyTangent;
	MyTangent.CalculateTangentSpace(myProxy);

	// fetch resulting data
	int32 iCount=0;
	for(f=0; f<p3dsObj->faces.size(); f++)
	{
		if(isFaceValid(p3dsObj->faces[f], p3dsObj))
		{
			DWORD indices[3];
			MyTangent.GetTriangleBaseIndices( iCount, indices );
			trngTang[f].i1 = indices[0];
			trngTang[f].i2 = indices[1];
			trngTang[f].i3 = indices[2];

			float outU[3], outV[3], outN[3];
			for(int q=0; q<3; q++)
			{
				MyTangent.GetBase(indices[q], outU, outV, outN);
				vS[indices[q]] = Vec3(outU[0], outU[1], outU[2]);
				vT[indices[q]] = Vec3(outV[0], outV[1], outV[2]);
			}

			// proceed
			++iCount;
		}
	}

	// Cleanup
	delete [] myProxy.m_pVerts;
	delete [] myProxy.m_pSTs;
	delete [] pNormals;
	delete [] myProxy.m_pPosIndices;
	delete [] myProxy.m_pSTIndices;
	delete [] myProxy.m_pNormIndices;
}

void CConv3DS::CreateGeometryData(CInternMesh * pIntMesh, C3DObject * p3dsObj, const Vec3 * pFinalNorms, const SInternFace * trngNrms, const Vec3 * pTangVectors, const Vec3 * pBinoVectors, const SInternFace * trngTang, const int32 ** descs, const int32 max_trng)
{
	// fetch geom data
	// into internal rep
	Vec3 * pVerts	= new Vec3[max_trng*3];	// tmp buffer
	Vec3 * pSTs		= new Vec3[max_trng*3];
	Vec3 * pNorms = new Vec3[max_trng*3];
	Vec3 * pTangs = new Vec3[max_trng*3];
	Vec3 * pBinos = new Vec3[max_trng*3];

	const int32 numSurfs = pIntMesh->m_iNumSurfs;
	
	for(int s=0; s<numSurfs; s++)
	{
		CInternSurface &surf = pIntMesh->m_pSurfaces[s];
		int32 uniquevertices = 0;
		
		for(int f=0; f<surf.m_iNumFaces; f++)
		{
			const int32 desc = (descs[s])[f];
			int32 indices[] = {p3dsObj->faces[desc].A, p3dsObj->faces[desc].B, p3dsObj->faces[desc].C};
			int32 indices_norms[] = {trngNrms[desc].i1, trngNrms[desc].i2, trngNrms[desc].i3};
			int32 indices_tangs[] = {trngTang[desc].i1, trngTang[desc].i2, trngTang[desc].i3};

			for(int i=0; i<3; i++)
			{
				// setup data
				const Vec3 vP(p3dsObj->vertices[indices[i]].x, p3dsObj->vertices[indices[i]].y, p3dsObj->vertices[indices[i]].z);
				const Vec3 vST = (p3dsObj->texturePoints.size()==0) ? Vec3(0,0,0) : Vec3(p3dsObj->texturePoints[indices[i]].x, p3dsObj->texturePoints[indices[i]].y, p3dsObj->texturePoints[indices[i]].z);
				const Vec3 vN = pFinalNorms[indices_norms[i]];
				const Vec3 vTang = pTangVectors[indices_tangs[i]];
				const Vec3 vBino = pBinoVectors[indices_tangs[i]];
				
				// look for a duplicate
				bool bFoundCpy = false;
				int k=0;
				while(k<uniquevertices && !bFoundCpy)
				{
					if(pVerts[k]==vP && pSTs[k]==vST && pNorms[k]==vN && pTangs[k]==vTang && pBinos[k]==vBino)
						bFoundCpy = true;
					else
						++k;	// increment
				}

				// if unique 
				if(!bFoundCpy)
				{
					k = uniquevertices;
					++uniquevertices;

					// deliver data to tmp buffer
					pVerts[k] = vP; pSTs[k] = vST; pNorms[k] = vN; pTangs[k] = vTang; pBinos[k]= vBino;
				}	

				// reordered index for the triangle indices
				indices[i] = k;
			}
			
			surf.m_pFaces[f] = SInternFace(indices[0], indices[1], indices[2]);
		}

		// deliver data to internal surface
		surf.m_iLenArrays = uniquevertices;
		surf.m_pVerts = new Vec3[surf.m_iLenArrays];
		surf.m_pSTs = new Vec3[surf.m_iLenArrays];
		surf.m_pNormals = new Vec3[surf.m_iLenArrays];
		surf.m_pTangents = new Vec3[surf.m_iLenArrays];
		surf.m_pBiNorms = new Vec3[surf.m_iLenArrays];
		for(int k=0; k<surf.m_iLenArrays; k++)
		{
			surf.m_pVerts[k] = pVerts[k];
			surf.m_pSTs[k] = pSTs[k];
			surf.m_pNormals[k] = pNorms[k];
			surf.m_pTangents[k] = pTangs[k];
			surf.m_pBiNorms[k] = pBinos[k];
		}
	}
	
	delete [] pVerts; delete [] pSTs; delete [] pNorms; delete [] pTangs; delete [] pBinos;	// remove tmp buffer
}


// Camera creation
void CConv3DS::MakeInternCamera(CInternCamera * pIntCam, const string name, const C3DCamera &c3dsCam, bool bFoundAnim)
{
	const Vec3 point = Vec3(c3dsCam.position.x, c3dsCam.position.y, c3dsCam.position.z);
	const float fFov = (float) (2400.0 / c3dsCam.lens);
	pIntCam->m_iRatioX = 4;
	pIntCam->m_iRatioY = 3;
	pIntCam->m_sCamName = new char[strlen(name.c_str())+1];
	strcpy(pIntCam->m_sCamName, name.c_str());

	// calculate transformation
	const Vec3 vTarget(c3dsCam.target.x, c3dsCam.target.y, c3dsCam.target.z);
	Mat33 m33 = CalcRot(point, vTarget, c3dsCam.bank);

	CreateTransform(pIntCam, Vec3(0,0,0), m33, point, bFoundAnim);
	m_pTmpAnim->SetFovKeyList();
	CreateKey1Lst(&pIntCam->m_FovArray, fFov, bFoundAnim);
}

void CConv3DS::MakeAllCameras(C3DModel &c3dsScene)
{
	// anims pointing to an existing camera count as an internal camera
	// cameras with no existing anim also count as an internal camera
	int32 max_cams = c3dsScene.GetCamNumber() + m_pTmpAnim->GetNumCameraAnims(), loc_cam;
	m_iNumCameras = 0;
	m_pCameras = new CInternCamera[max_cams];	
	for(int c=0; c<m_pTmpAnim->GetNumCameraAnims(); c++)
		//	anim_has_camera()
		if(m_pTmpAnim->SetCamera(c) && 
			(loc_cam=GetCameraIndex(c3dsScene, m_pTmpAnim->GetName())) != -1)
		{
			const string name = m_pTmpAnim->GetTrackInstName().empty()?m_pTmpAnim->GetName():m_pTmpAnim->GetTrackInstName();
			MakeInternCamera(&m_pCameras[m_iNumCameras++], name, c3dsScene.GetCamera(loc_cam), true);
		}
	for(c=0; c<c3dsScene.GetCamNumber(); c++)
		// camera_has_no_anim()
		if(!m_pTmpAnim->SetCamera( c3dsScene.GetCamera(c).m_sCamName ) )
		{
			const string name = c3dsScene.GetCamera(c).m_sCamName;
			MakeInternCamera(&m_pCameras[m_iNumCameras++], name, c3dsScene.GetCamera(c), false);
		}

	assert(m_iNumCameras <= max_cams);
}


// Light creation
void CConv3DS::MakeInternLight(CInternLight * pIntLgt, const string name, const C3DLight &c3dsLgt, bool bFoundAnim)
{
	const Vec3 point = Vec3(c3dsLgt.position.x, c3dsLgt.position.y, c3dsLgt.position.z);
	const Vec3 color = Vec3(c3dsLgt.r, c3dsLgt.g, c3dsLgt.b);
	const float fNear = c3dsLgt.Near;
	const float fFar = c3dsLgt.Far;
	
	pIntLgt->m_sLightName = new char[strlen(name.c_str())+1];
	strcpy(pIntLgt->m_sLightName, name.c_str());
	pIntLgt->m_bIsSpot = c3dsLgt.is_spot;
	pIntLgt->m_bIsRectangular = c3dsLgt.is_rectangular;
	
	const float fHotSpot = c3dsLgt.hotspot;
	const float fFallOff = c3dsLgt.falloff;
	const float fAspect = c3dsLgt.aspect;

	// calculate transformation
	const Vec3 vTarget(c3dsLgt.target.x, c3dsLgt.target.y, c3dsLgt.target.z);
	Mat33 m33 = CalcRot(point, vTarget, c3dsLgt.roll);
	if(!c3dsLgt.is_spot)
		LoadIdentity(&m33);

	// setup splines
	CreateTransform(pIntLgt, Vec3(0,0,0), m33, point, bFoundAnim);
	CreateKey1Lst(&pIntLgt->m_NearArray, fNear, false);
	CreateKey1Lst(&pIntLgt->m_FarArray, fFar, false);
	CreateKey3Lst(&pIntLgt->m_ColorArray, color, false);

	bFoundAnim = bFoundAnim && pIntLgt->IsSpot();
	m_pTmpAnim->SetHotSpotKeyList();
	CreateKey1Lst(&pIntLgt->m_HotSpotArray, fHotSpot, bFoundAnim);
	m_pTmpAnim->SetFallOffKeyList();
	CreateKey1Lst(&pIntLgt->m_FallOffArray, fFallOff, bFoundAnim);
	CreateKey1Lst(&pIntLgt->m_AspectArray, fAspect, false);
}

void CConv3DS::MakeAllLights(C3DModel &c3dsScene)
{
	// anims pointing to an existing light count as an internal light
	// cameras with no existing anim also count as an internal light
	int32 max_lgts = c3dsScene.GetLightNumber() + m_pTmpAnim->GetNumOmniAnims() + m_pTmpAnim->GetNumSpotAnims(), loc_lgt;
	m_iNumLights = 0;
	m_pLights = new CInternLight[max_lgts];	
	for(int l=0; l<m_pTmpAnim->GetNumOmniAnims(); l++)
		//	anim_has_omni()
		if(m_pTmpAnim->SetOmni(l) && 
			(loc_lgt=GetOmniIndex(c3dsScene, m_pTmpAnim->GetName())) != -1)
		{
			const string name = m_pTmpAnim->GetTrackInstName().empty()?m_pTmpAnim->GetName():m_pTmpAnim->GetTrackInstName();
			MakeInternLight(&m_pLights[m_iNumLights++], name, c3dsScene.GetLight(loc_lgt), true);
		}
	for(l=0; l<m_pTmpAnim->GetNumSpotAnims(); l++)
		//	anim_has_spot()
		if(m_pTmpAnim->SetSpot(l) && 
			(loc_lgt=GetSpotIndex(c3dsScene, m_pTmpAnim->GetName())) != -1)
		{
			const string name = m_pTmpAnim->GetTrackInstName().empty()?m_pTmpAnim->GetName():m_pTmpAnim->GetTrackInstName();
			MakeInternLight(&m_pLights[m_iNumLights++], name, c3dsScene.GetLight(loc_lgt), true);
		}
	for(l=0; l<c3dsScene.GetLightNumber(); l++)
	{
		const C3DLight &c3dsLgt = c3dsScene.GetLight(l);
		bool bFoundAnim = (c3dsLgt.is_spot) ? m_pTmpAnim->SetSpot( c3dsLgt.m_sLightName ) : m_pTmpAnim->SetOmni( c3dsLgt.m_sLightName );
		// light_has_no_anim()
		if(!bFoundAnim)
		{
			const string name = c3dsLgt.m_sLightName;
			MakeInternLight(&m_pLights[m_iNumLights++], name, c3dsLgt, false);
		}
	}

	assert(m_iNumLights <= max_lgts);
}


// 
const int32 CConv3DS::Max(const int32 a, const int32 b) { return (a>b) ? a : b; }

void CConv3DS::CreateTransform(CTransformation * pTrans, const Vec3 &vCen, const Mat33 &mDefault33, const Vec3 &vDefault3, const bool bAnimFound)
{
	int32 num_pos_keys = 0;
	int32 num_rot_keys = 0;
	int32 num_scl_keys = 0;
	int32 num_dst_pos_keys = 1;

	// determine what kind of ressources we have in quantity
	// and also decide how many translation points to output
	if(bAnimFound)
	{
		m_pTmpAnim->SetPositionKeyList();
		num_pos_keys = m_pTmpAnim->GetNumKey3s();
		m_pTmpAnim->SetRotationKeyList();
		num_rot_keys = m_pTmpAnim->GetNumMat33s();
		m_pTmpAnim->SetScaleKeyList();
		num_scl_keys = m_pTmpAnim->GetNumKey1s();

		// having a center will affect num of translation samples since rotation
		// and scale may have more samples then the translation part and the
		// center will be affected by those.
		if(vCen != Vec3(0,0,0))
			num_dst_pos_keys = Max(Max(num_dst_pos_keys, num_scl_keys), Max(num_pos_keys, num_rot_keys));
		else
			num_dst_pos_keys = Max(num_dst_pos_keys, num_pos_keys);
	}

	// if we have animated source points
	pTrans->m_PosArray.m_iNumKeys = num_dst_pos_keys;
	pTrans->m_PosArray.m_pKeys = new Vec3[pTrans->GetNumPosKeys()];
	m_pTmpAnim->SetPositionKeyList();
	if(bAnimFound && num_pos_keys>0)
		for(int t=0; t<num_dst_pos_keys; t++)
		{
			float frame = (num_dst_pos_keys>1) ? (t*(num_pos_keys-1.0f))/(num_dst_pos_keys-1) : 0; 
			pTrans->m_PosArray.m_pKeys[t] = m_pTmpAnim->GetKey3(frame);
		}
	else
		for(int t=0; t<num_dst_pos_keys; t++)
			pTrans->m_PosArray.m_pKeys[t] = vDefault3;

	// pick up scale values
	pTrans->m_ScaleArray.m_iNumKeys = (num_scl_keys>0)?num_scl_keys:1;
	pTrans->m_ScaleArray.m_pKeys = new float[pTrans->GetNumScaleKeys()];
	m_pTmpAnim->SetScaleKeyList();

	if(bAnimFound && num_scl_keys>0)
		for(int s=0; s<num_scl_keys; s++)
			pTrans->m_ScaleArray.m_pKeys[s] = m_pTmpAnim->GetKey1(s);
	else
		pTrans->m_ScaleArray.m_pKeys[0] = 1.0f;

	// pick up rotations
	pTrans->m_RotArray.m_iNumKeys = (num_rot_keys>0)?num_rot_keys:1; 
	pTrans->m_RotArray.m_pKeys = new Quat[pTrans->GetNumRotKeys()];
	m_pTmpAnim->SetRotationKeyList();
	const Mat33 mAxisUp = GetAxisUpMatrix();
	
	if(bAnimFound && num_rot_keys>0)
		for(int r=0; r<num_rot_keys; r++)
			pTrans->m_RotArray.m_pKeys[r] = RotationToQuat(mAxisUp * m_pTmpAnim->GetMat33(r) * Transpose(mAxisUp));
	else
		pTrans->m_RotArray.m_pKeys[0] = RotationToQuat(mAxisUp * mDefault33 * Transpose(mAxisUp));

	// last manipulation of translation vectors
	// pTrans, vCen, m33, p3
	// apply new center and coordinate system conversion matrix
	// mAxisUpMat * T * R * S * cen * Transpose( mAxisUpMat )		"actual transform of matrices"
	// 1. T += (S * R*cen)														"modified fix"
	// 2. T = mAxisUp * T
	// 3. R = (mAxisUp * R) * Transpose( mAxisUpMat )

	if(vCen != Vec3(0,0,0))
		for(int t=0; t<num_dst_pos_keys; t++)
		{
			Mat33 rot;
			if(bAnimFound && num_rot_keys>0)
			{
				float frameR = (num_dst_pos_keys>1) ? (t*(num_rot_keys-1.0f))/(num_dst_pos_keys-1) : 0; 
				rot = m_pTmpAnim->GetMat33(frameR);
			}
			else
				rot = mDefault33;

			
			float scale = 1.0f;	// default value
			if(bAnimFound && num_scl_keys>0)
			{
				float frameS = (num_dst_pos_keys>1) ? (t*(num_scl_keys-1.0f))/(num_dst_pos_keys-1) : 0; 
				scale = pTrans->GetScaleKey(frameS);
			}
			
			pTrans->m_PosArray.m_pKeys[t] += (scale * (rot * vCen));
			pTrans->m_PosArray.m_pKeys[t] = mAxisUp * pTrans->m_PosArray.m_pKeys[t];
		}
	else
		for(int t=0; t<num_dst_pos_keys; t++)
			pTrans->m_PosArray.m_pKeys[t] = mAxisUp * pTrans->m_PosArray.m_pKeys[t];	
}

void CConv3DS::CreateKey1Lst(CInternKey1List * pKey1Lst, const float fDefault, const bool bAnimFound)
{
	// pick up spline from keyframe chunk
	if(bAnimFound && m_pTmpAnim->GetNumKey1s()>0)
	{
		pKey1Lst->m_iNumKeys	= m_pTmpAnim->GetNumKey1s();
		pKey1Lst->m_pKeys = new float[pKey1Lst->m_iNumKeys];
		for(int k=0; k<pKey1Lst->m_iNumKeys; k++)
			pKey1Lst->m_pKeys[k] = m_pTmpAnim->GetKey1(k);
	}
	else	// use default value
	{
		pKey1Lst->m_iNumKeys = 1;
		pKey1Lst->m_pKeys = new float;
		pKey1Lst->m_pKeys[0] = fDefault;
	}
}

void CConv3DS::CreateKey3Lst(CInternKey3List * pKey3Lst, const Vec3 &vDefault, const bool bAnimFound)
{
	// pick up spline from keyframe chunk
	if(bAnimFound && m_pTmpAnim->GetNumKey3s()>0)
	{
		pKey3Lst->m_iNumKeys	= m_pTmpAnim->GetNumKey3s();
		pKey3Lst->m_pKeys = new Vec3[pKey3Lst->m_iNumKeys];
		for(int k=0; k<pKey3Lst->m_iNumKeys; k++)
			pKey3Lst->m_pKeys[k] = m_pTmpAnim->GetKey3(k);
	}
	else	// use default value
	{
		pKey3Lst->m_iNumKeys = 1;
		pKey3Lst->m_pKeys = new Vec3;
		pKey3Lst->m_pKeys[0] = vDefault;
	}
}



// Little utility function
const Mat33 CConv3DS::CalcRot(const Vec3 &vPos, const Vec3 &vTarget, const float roll)
{
	// calculate lookAt matrix
	Vec3 c1(1,0,0), c2(0,1,0), c3(0,0,1);
	Vec3 vDir = vTarget - vPos;
	if(vDir != Vec3(0,0,0))
	{
		vDir = Normalize(vDir);
		if(fabs(vDir*Vec3(0,0,1)) < 1)
			c1 = Normalize( Cross(vDir, Vec3(0,0,1)) );
		c2 = vDir;
		c3 = Normalize( Cross(c1, c2) );
	}

	// make result
	Mat33 m33;
	SetColumn(&m33, 0, c1);
	SetColumn(&m33, 1, c2);
	SetColumn(&m33, 2, c3);
	if(roll != 0)
	{
		// roll around nose
		Mat33 r33;
		const float fAngToRad = (float) (3.1415926535897932384626433832795 / 180);
		LoadRotationAxisAngle(&r33, GetColumn(m33, 1), roll * fAngToRad);
		m33 = r33 * m33;
	}
	
	// return result
	return m33;
}

const Mat33 CConv3DS::GetAxisUpMatrix()
{
	Mat33 mAxisUp;
	if(m_bForceYaxisUp)
	{
		SetColumn(&mAxisUp, 0, Vec3(1,0,0));
		SetColumn(&mAxisUp, 1, Vec3(0,0,-1));
		SetColumn(&mAxisUp, 2, Vec3(0,1,0));
	}
	else
		LoadIdentity(&mAxisUp);

	return mAxisUp;
}


// anim finder
int32 CConv3DS::GetMeshIndex(C3DModel &c3dsScene, string Name)
{
	int32 res = -1;

	int m=0;
	while(m<c3dsScene.GetObjNumber() && res==-1)
		if(c3dsScene.GetObject(m)->m_sObjName == Name)
			res = m;
		else
			++m;
	
	return res;
}

int32 CConv3DS::GetCameraIndex(C3DModel &c3dsScene, string Name)
{
	int32 res = -1;

	int c=0;
	while(c<c3dsScene.GetCamNumber() && res==-1)
		if(c3dsScene.GetCamera(c).m_sCamName == Name)
			res = c;
		else
			++c;
	
	return res;
}

int32 CConv3DS::GetOmniIndex(C3DModel &c3dsScene, string Name)
{
	int32 res = -1;

	int l=0;
	while(l<c3dsScene.GetLightNumber() && res==-1)
		if(c3dsScene.GetLight(l).m_sLightName == Name && !c3dsScene.GetLight(l).is_spot)
			res = l;
		else
			++l;
	
	return res;
}

int32 CConv3DS::GetSpotIndex(C3DModel &c3dsScene, string Name)
{
	int32 res = -1;

	int l=0;
	while(l<c3dsScene.GetLightNumber() && res==-1)
		if(c3dsScene.GetLight(l).m_sLightName == Name && c3dsScene.GetLight(l).is_spot)
			res = l;
		else
			++l;
	
	return res;
}
