#include "convtempanim.h"
#include "C3DModel.h"
#include "tcbctrl.h"
#include <assert.h>


	
void CConvTempAnim::Setup(const C3DModel &c3dsScene)
{
	m_iFrameBegin = c3dsScene.GetFrameBegin();
	m_iFrameEnd = c3dsScene.GetFrameEnd();

	// alloc working data
	Key * pKeys = new Key[(m_iFrameEnd-m_iFrameBegin)+1];
	m_pAnimCtrl = new CTCBAnimCtrl;
	m_pAnimCtrl->SetAnimInterval(m_iFrameBegin, m_iFrameEnd);
	m_pAnimCtrl->SetAnimKeys(pKeys, 1);	// default num_keys is 1
	
	// prepare
	m_iNumMeshAnims = c3dsScene.GetNumMeshAnims();
	m_iNumCamAnims = c3dsScene.GetNumCameraAnims();
	m_iNumOmniAnims = c3dsScene.GetNumOmniAnims();
	m_iNumSpotAnims = c3dsScene.GetNumSpotAnims();
	m_iNumCamTargetAnims = c3dsScene.GetNumCameraTargetAnims();
	m_iNumSpotTargetAnims = c3dsScene.GetNumSpotTargetAnims();
	
	// setup new ressources
	m_pTmpMeshAnim = (m_iNumMeshAnims>0) ? new CTempMeshAnim[m_iNumMeshAnims] : NULL;
	m_pTmpCamAnim = (m_iNumCamAnims>0) ? new CTempCamAnim[m_iNumCamAnims] : NULL;
	m_pTmpOmniAnim = (m_iNumOmniAnims>0) ? new CTempOmniAnim[m_iNumOmniAnims] : NULL;
	m_pTmpSpotAnim = (m_iNumSpotAnims>0) ? new CTempSpotAnim[m_iNumSpotAnims] : NULL;
	m_pTmpCamTargetAnim = (m_iNumCamTargetAnims>0) ? new CTempCamTgtAnim[m_iNumCamTargetAnims] : NULL;
	m_pTmpSpotTargetAnim = (m_iNumSpotTargetAnims>0) ? new CTempSpotTgtAnim[m_iNumSpotTargetAnims] : NULL;

	
	// convert all meshes
	for(int32 m=0; m<m_iNumMeshAnims; m++)
		Do3dsToTempMeshAnim(&m_pTmpMeshAnim[m], c3dsScene.GetTrackInfo(m));

	// convert all cameras
	for(int32 c=0; c<m_iNumCamAnims; c++)
		Do3dsToTempCamAnim(&m_pTmpCamAnim[c], c3dsScene.GetTrackCamera(c));
	
	// convert all omnis
	for(int32 o=0; o<m_iNumOmniAnims; o++)
		Do3dsToTempOmniAnim(&m_pTmpOmniAnim[o], c3dsScene.GetTrackOmni(o));

	// convert all spots
	for(int32 s=0; s<m_iNumSpotAnims; s++)
		Do3dsToTempSpotAnim(&m_pTmpSpotAnim[s], c3dsScene.GetTrackSpot(s));

	// convert all camera targets
	for(c=0; c<m_iNumCamTargetAnims; c++)
		Do3dsToTempCamTargetAnim(&m_pTmpCamTargetAnim[c], c3dsScene.GetTrackCamTarget(c));
	
	// convert all spot targets
	for(s=0; s<m_iNumSpotTargetAnims; s++)
		Do3dsToTempSpotTargetAnim(&m_pTmpSpotTargetAnim[s], c3dsScene.GetTrackSpotTarget(s));

	// dealloc working data
	delete [] pKeys;
	delete m_pAnimCtrl;

	// setup hierarchy
	const int32 iTotNumNodesInHierarchy = m_iNumMeshAnims + m_iNumCamAnims + m_iNumOmniAnims + m_iNumSpotAnims + m_iNumCamTargetAnims + m_iNumSpotTargetAnims;
	CTempTransform ** ppTmpTrans = (iTotNumNodesInHierarchy>0) ?  new CTempTransform *[iTotNumNodesInHierarchy] : NULL;
	int index=0;
	for(m=0; m<m_iNumMeshAnims; m++) ppTmpTrans[index++] = &m_pTmpMeshAnim[m];
	for(c=0; c<m_iNumCamAnims; c++) ppTmpTrans[index++] = &m_pTmpCamAnim[c];
	for(o=0; o<m_iNumOmniAnims; o++) ppTmpTrans[index++] = &m_pTmpOmniAnim[o];
	for(s=0; s<m_iNumSpotAnims; s++) ppTmpTrans[index++] = &m_pTmpSpotAnim[s];
	for(c=0; c<m_iNumCamTargetAnims; c++) ppTmpTrans[index++] = &m_pTmpCamTargetAnim[c];
	for(s=0; s<m_iNumSpotTargetAnims; s++) ppTmpTrans[index++] = &m_pTmpSpotTargetAnim[s];
	HandleHierarchy(ppTmpTrans, iTotNumNodesInHierarchy);
	if(ppTmpTrans != NULL)
		delete [] ppTmpTrans;

	// fix direct camera and spot transformations now that
	// all have been transformed to world coordinate system
	for(c=0; c<m_iNumCamAnims; c++)
		FixupTransDirectional(&m_pTmpCamAnim[c], m_pTmpCamAnim[c].m_cRoll, m_pTmpCamTargetAnim, m_iNumCamTargetAnims);
	for(s=0; s<m_iNumSpotAnims; s++)
		FixupTransDirectional(&m_pTmpSpotAnim[s], m_pTmpSpotAnim[s].m_cRoll, m_pTmpSpotTargetAnim, m_iNumSpotTargetAnims);


	// trim precission of splines
	for(m=0; m<m_iNumMeshAnims; m++)
		DownSampleMeshSplines(&m_pTmpMeshAnim[m]);

	for(c=0; c<m_iNumCamAnims; c++)
		DownSampleCameraSplines(&m_pTmpCamAnim[c]);
	
	for(o=0; o<m_iNumOmniAnims; o++)
		DownSampleOmniSplines(&m_pTmpOmniAnim[o]);

	for(s=0; s<m_iNumSpotAnims; s++)
		DownSampleSpotSplines(&m_pTmpSpotAnim[s]);

	for(c=0; c<m_iNumCamTargetAnims; c++)
		DeleteCamTargetSplines(&m_pTmpCamTargetAnim[c]);
	if(m_pTmpCamTargetAnim!=NULL)
		delete [] m_pTmpCamTargetAnim;
	
	for(s=0; s<m_iNumSpotTargetAnims; s++)
		DeleteSpotTargetSplines(&m_pTmpSpotTargetAnim[s]);
	if(m_pTmpSpotTargetAnim!=NULL)
		delete [] m_pTmpSpotTargetAnim;
}
	
void CConvTempAnim::Do3dsToTempMeshAnim(CTempMeshAnim * pMeshAnim, const C3DTrackInfo * pTrackInfo)
{
	Do3dsToTempTransAnim((CTempTransform *) pMeshAnim, (C3DTrackInfo *) pTrackInfo);
}

void CConvTempAnim::Do3dsToTempCamAnim(CTempCamAnim * pCamAnim, const C3DTrackCamera * pTrackCam)
{
	Do3dsToTempTransAnim((CTempTransform *) pCamAnim, (C3DTrackInfo *) pTrackCam);
	const int32 num_keys = (m_iFrameEnd - m_iFrameBegin) + 1;

	// pickup roll	
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTrackCam->GetNumRollKeys());
	for(int k=0; k<pTrackCam->GetNumRollKeys(); k++)
	{
		const keyf1 &key = pTrackCam->GetRollKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.x, key.x, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pCamAnim->m_cRoll.m_iNumKeys = num_keys;
	pCamAnim->m_cRoll.m_pKeys = new float[num_keys];
	for(int i=0; i<num_keys; i++)
		if(pTrackCam->GetNumRollKeys()>0)
			pCamAnim->m_cRoll.m_pKeys[i] = GetUniScale(i + m_iFrameBegin, pTrackCam->GetNumRollKeys());
		else
			pCamAnim->m_cRoll.m_pKeys[i] = 0;

	// pickup fov
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTrackCam->GetNumFovKeys());
	for(k=0; k<pTrackCam->GetNumFovKeys(); k++)
	{
		const keyf1 &key = pTrackCam->GetFovKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.x, key.x, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pCamAnim->m_cFov.m_iNumKeys = num_keys;
	pCamAnim->m_cFov.m_pKeys = new float[num_keys];
	for(i=0; i<num_keys; i++)
		if(pTrackCam->GetNumFovKeys()>0)
			pCamAnim->m_cFov.m_pKeys[i] = GetUniScale(i + m_iFrameBegin, pTrackCam->GetNumFovKeys());
		else
			pCamAnim->m_cFov.m_pKeys[i] = 45;
}

void CConvTempAnim::Do3dsToTempOmniAnim(CTempOmniAnim * pOmniAnim, const C3DTrackOmni * pTrackOmni)
{
	Do3dsToTempTransAnim((CTempTransform *) pOmniAnim, (C3DTrackInfo *) pTrackOmni);
}

void CConvTempAnim::Do3dsToTempSpotAnim(CTempSpotAnim * pSpotAnim, const C3DTrackSpot * pTrackSpot)
{
	Do3dsToTempTransAnim((CTempTransform *) pSpotAnim, (C3DTrackInfo *) pTrackSpot);
	const int32 num_keys = (m_iFrameEnd - m_iFrameBegin) + 1;
	
	// pickup roll
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTrackSpot->GetNumRollKeys());
	for(int k=0; k<pTrackSpot->GetNumRollKeys(); k++)
	{
		const keyf1 &key = pTrackSpot->GetRollKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.x, key.x, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pSpotAnim->m_cRoll.m_iNumKeys = num_keys;
	pSpotAnim->m_cRoll.m_pKeys = new float[num_keys];
	for(int i=0; i<num_keys; i++)
		if(pTrackSpot->GetNumRollKeys()>0)
			pSpotAnim->m_cRoll.m_pKeys[i] = GetUniScale(i + m_iFrameBegin, pTrackSpot->GetNumRollKeys());
		else
			pSpotAnim->m_cRoll.m_pKeys[i] = 0;

	// pickup hotspot
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTrackSpot->GetNumHotSpotKeys());
	for(k=0; k<pTrackSpot->GetNumHotSpotKeys(); k++)
	{
		const keyf1 &key = pTrackSpot->GetHotSpotKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.x, key.x, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pSpotAnim->m_cHotSpot.m_iNumKeys = num_keys;
	pSpotAnim->m_cHotSpot.m_pKeys = new float[num_keys];
	for(i=0; i<num_keys; i++)
		if(pTrackSpot->GetNumHotSpotKeys()>0)
			pSpotAnim->m_cHotSpot.m_pKeys[i] = GetUniScale(i + m_iFrameBegin, pTrackSpot->GetNumHotSpotKeys());
		else
			pSpotAnim->m_cHotSpot.m_pKeys[i] = 43;

	// pickup falloff
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTrackSpot->GetNumFallOffKeys());
	for(k=0; k<pTrackSpot->GetNumFallOffKeys(); k++)
	{
		const keyf1 &key = pTrackSpot->GetFallOffKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.x, key.x, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pSpotAnim->m_cFallOff.m_iNumKeys = num_keys;
	pSpotAnim->m_cFallOff.m_pKeys = new float[num_keys];
	for(i=0; i<num_keys; i++)
		if(pTrackSpot->GetNumFallOffKeys()>0)
			pSpotAnim->m_cFallOff.m_pKeys[i] = GetUniScale(i + m_iFrameBegin, pTrackSpot->GetNumFallOffKeys());
		else
			pSpotAnim->m_cFallOff.m_pKeys[i] = 45;
}

void CConvTempAnim::Do3dsToTempCamTargetAnim(CTempCamTgtAnim * pCamTargetAnim, const C3DTrackCamTarget * pTrackCamTarget)
{
	Do3dsToTempTransAnim((CTempTransform *) pCamTargetAnim, (C3DTrackInfo *) pTrackCamTarget);
}

void CConvTempAnim::Do3dsToTempSpotTargetAnim(CTempSpotTgtAnim * pSpotTargetAnim, const C3DTrackSpotTarget * pTrackSpotTarget)
{
	Do3dsToTempTransAnim((CTempTransform *) pSpotTargetAnim, (C3DTrackInfo *) pTrackSpotTarget);
}

void CConvTempAnim::Do3dsToTempTransAnim(CTempTransform * pTransOut, const C3DTrackInfo * pTransIn)
{
	const int32 num_keys = (m_iFrameEnd - m_iFrameBegin) + 1;
	pTransOut->m_iHierarchyNum = pTransIn->GetHierarchyNum();
	pTransOut->m_iNodeID = pTransIn->GetNodeID(); 
	pTransOut->m_s3dsName = pTransIn->GetNameID();
	pTransOut->m_s3dsTrackInstName = pTransIn->GetTrackInstName();

	const Vec3 pivot(pTransIn->GetPivotPoint().x, pTransIn->GetPivotPoint().y, pTransIn->GetPivotPoint().z);

	// setup positions
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTransIn->GetNumPosKeys());
	for(int k=0; k<pTransIn->GetNumPosKeys(); k++)
	{
		const keyf3 &key = pTransIn->GetPosKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.y, key.z, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pTransOut->m_cTranslation.m_iNumKeys = num_keys;
	pTransOut->m_cTranslation.m_pKeys = new Vec3[num_keys];
	for(int i=0; i<num_keys; i++)
		if(pTransIn->GetNumPosKeys()>0)
			pTransOut->m_cTranslation.m_pKeys[i] = GetPosition(i + m_iFrameBegin, pTransIn->GetNumPosKeys());
		else
			pTransOut->m_cTranslation.m_pKeys[i] = Vec3(0,0,0);
	
	// setup rotations
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTransIn->GetNumRotKeys());
	for(k=0; k<pTransIn->GetNumRotKeys(); k++)
	{
		const keyf4 &key = pTransIn->GetRotKey(k);
		m_pAnimCtrl->SetRotKey(k, key.key_num, key.x, key.y, key.z, key.w, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}
	m_pAnimCtrl->UpdateQuatDerivates();
	pTransOut->m_cRotation.m_iNumKeys = num_keys;
	pTransOut->m_cRotation.m_pKeys = new Mat33[num_keys];
	for(int r=0; r<num_keys; r++)
		if(pTransIn->GetNumRotKeys()>0)
			pTransOut->m_cRotation.m_pKeys[r] = GetRotation(r + m_iFrameBegin, pTransIn->GetNumRotKeys());
		else
			LoadIdentity(&pTransOut->m_cRotation.m_pKeys[r]);	
	
	// Scale
	m_pAnimCtrl->SetAnimKeys(m_pAnimCtrl->GetKeys(), pTransIn->GetNumScaleKeys());
	for(k=0; k<pTransIn->GetNumScaleKeys(); k++)
	{
		const keyf3 &key = pTransIn->GetScaleKey(k);
		m_pAnimCtrl->SetKey(k, key.key_num, key.x, key.y, key.z, key.tens, key.cont, key.bias, key.easeto, key.easefrom);
	}	
	m_pAnimCtrl->UpdateDerivates();
	pTransOut->m_cUniScale.m_iNumKeys = num_keys;
	pTransOut->m_cUniScale.m_pKeys = new float[num_keys];
	for(int s=0; s<num_keys; s++)
		if(pTransIn->GetNumScaleKeys()>0)
			pTransOut->m_cUniScale.m_pKeys[s] = GetUniScale(s + m_iFrameBegin, pTransIn->GetNumScaleKeys());
		else
			pTransOut->m_cUniScale.m_pKeys[s] = 1;

	// apply pivot point if necessary
	if(pivot != Vec3(0,0,0))
		for(int k=0; k<num_keys; k++)
		{
			// in matrix form the order is Piv * Trans * Rot * UniScale * (-Piv)
			// if we let vPiv, vTrans be vectors and fUniScale be a float then we get
			// vNewTrans =  /*vPiv +*/ vTrans + fUniScale * Rot * (-vPiv)
			// the new order of opertations then becomes
			// NewTrans * Rot * UniScale
			Vec3 tmp_vert = pTransOut->m_cRotation.m_pKeys[k]*pivot;
			tmp_vert *= (-pTransOut->m_cUniScale.m_pKeys[k]);
			pTransOut->m_cTranslation.m_pKeys[k] += (/*pivot + */tmp_vert);
		}
}

const Vec3 CConvTempAnim::GetPosition(const float t, const int num_keys)
{
	Vec3 vRes;

	m_pAnimCtrl->EvalPos(t, &vRes);
	
	return vRes;
}

const Mat33 CConvTempAnim::GetRotation(const float t, const int num_keys)
{
	Mat33 m33;
	Quat q;
	m_pAnimCtrl->EvalRot(t, &q);
	LoadRotation(&m33, q);

	return Transpose(m33);	// for some reason they're stored as inverse
}

const float CConvTempAnim::GetUniScale(const float t, const int num_keys)
{
	Vec3 scale;
	m_pAnimCtrl->EvalPos(t, &scale);

	return (scale.x+scale.y+scale.z)/3;
}

void CConvTempAnim::HandleHierarchy(CTempTransform ** ppTmpTrans, const int32 iTotNumNodesInHierarchy)
{
	for(int k=0; k<iTotNumNodesInHierarchy; k++)
	{
		bool bNotFound = true;
		int p=0;
		while(p<iTotNumNodesInHierarchy && bNotFound)
		{
			if(ppTmpTrans[k]->m_iHierarchyNum == ppTmpTrans[p]->m_iNodeID)
			{
				// Parent ID must be valid!
				assert(ppTmpTrans[p]->m_iNodeID>=0 && k!=p);

				// insert child
				ppTmpTrans[k]->m_pParent = ppTmpTrans[p];
				if(ppTmpTrans[p]->m_pFirstChild == NULL)
				{
					ppTmpTrans[p]->m_pFirstChild = ppTmpTrans[k];
					assert(ppTmpTrans[k]->m_pNextSibling == NULL);
				}
				else
				{
					CTempTransform * pChildNext = ppTmpTrans[p]->m_pFirstChild;
					while(pChildNext->m_pNextSibling != NULL)
						pChildNext = pChildNext->m_pNextSibling;
					pChildNext->m_pNextSibling = ppTmpTrans[k];
				}

				bNotFound = false;
			}
			else ++p;
		}
		
		if(ppTmpTrans[k]->m_iHierarchyNum>=0 && bNotFound)
			printf("warning! parent node not found for %s\n", ppTmpTrans[k]->m_s3dsTrackInstName.c_str());
	}

	// transform to world
	for(k=0; k<iTotNumNodesInHierarchy; k++)
	{
		// if no parent
		if(ppTmpTrans[k]->m_iHierarchyNum<0)
		{
			assert(ppTmpTrans[k]->m_pNextSibling == NULL);
			ApplyRecurTransform(ppTmpTrans[k]);
		}
	}

	/*
	// sort by nodeID
	for(int k=0; k<iTotNumNodesInHierarchy-1; k++)
		for(int m=k+1; m<iTotNumNodesInHierarchy; m++)
			if(ppTmpTrans[k]->m_iNodeID > ppTmpTrans[m]->m_iNodeID)
			{
				CTempTransform * pTmp = ppTmpTrans[m];
				ppTmpTrans[m] = ppTmpTrans[k];
				ppTmpTrans[k] = pTmp;
			}

	// setup tree
	CTempTransform superNode;
	superNode.m_iHierarchyNum = -2;	// less than anyone else
	CTempTransform * pCurParent = &superNode;

	// distribute all nodes
	for(int h=0; h<iTotNumNodesInHierarchy; h++)
	{
		CTempTransform * pNodeInsert = ppTmpTrans[h];
		assert(pNodeInsert->m_iHierarchyNum > -2);
		while(pNodeInsert->m_iHierarchyNum <= pCurParent->m_iHierarchyNum)
		{
			pCurParent = pCurParent->m_pParent;
			assert(pCurParent!=NULL);
		}

		// insert child
		pNodeInsert->m_pParent = pCurParent;
		if(pCurParent->m_pFirstChild == NULL)
		{
			pCurParent->m_pFirstChild = pNodeInsert;
			assert(pNodeInsert->m_pNextSibling == NULL);
		}
		else
		{
			CTempTransform	* pChildNext = pCurParent->m_pFirstChild;
			while(pChildNext->m_pNextSibling != NULL)
				pChildNext = pChildNext->m_pNextSibling;
			pChildNext->m_pNextSibling = pNodeInsert;
		}

		// mark ourselves as the new potential parent
		pCurParent = pNodeInsert;
	}

	// transform to world
	CTempTransform	* pChildNext = superNode.m_pFirstChild;
	while(pChildNext!=NULL)
	{
		ApplyRecurTransform(pChildNext);
		pChildNext = pChildNext->m_pNextSibling;
	}*/
}

void CConvTempAnim::ApplyRecurTransform(CTempTransform * pCurNode)
{
	// transform to world
	const int32 num_keys = (m_iFrameEnd - m_iFrameBegin) + 1;
	CTempTransform	* pChildNext = pCurNode->m_pFirstChild;
	while(pChildNext!=NULL)
	{
		// pCurNode represents the 2's
		// pChildNext is obviously the 1's
		// mToWorld = (T2*R2*S2) * (T1*R1*S1)		"actual transformation"
		//  T1 = T2 + S2*(R2*T1)		"semi clever substitution"
		//  S1 *= S2	(uniscale)
		//	 R1 = R2*R1
		assert(pChildNext->m_cTranslation.GetNumKeys() == num_keys);
		assert(pChildNext->m_cRotation.GetNumKeys() == num_keys);
		assert(pChildNext->m_cUniScale.GetNumKeys() == num_keys);
		assert(pCurNode->m_cTranslation.GetNumKeys() == num_keys);
		assert(pCurNode->m_cRotation.GetNumKeys() == num_keys);
		assert(pCurNode->m_cUniScale.GetNumKeys() == num_keys);

		for(int k=0; k<num_keys; k++)
		{
			Vec3 * pPos1 = &pChildNext->m_cTranslation.m_pKeys[k];
			Mat33 * pRot1 = &pChildNext->m_cRotation.m_pKeys[k];
			float * pScale1 = &pChildNext->m_cUniScale.m_pKeys[k];
			Vec3 * pPos2 = &pCurNode->m_cTranslation.m_pKeys[k];
			Mat33 * pRot2 = &pCurNode->m_cRotation.m_pKeys[k];
			float * pScale2 = &pCurNode->m_cUniScale.m_pKeys[k];

			// transformed to world
			pPos1[0] = pPos2[0] + pScale2[0]*(pRot2[0]*pPos1[0]);
			pScale1[0] *= pScale2[0];
			pRot1[0] = pRot2[0]*pRot1[0];
		}
		
		pChildNext = pChildNext->m_pNextSibling;
	}

	// recurse
	pChildNext = pCurNode->m_pFirstChild;
	while(pChildNext!=NULL)
	{
		ApplyRecurTransform(pChildNext);
		pChildNext = pChildNext->m_pNextSibling;
	}
}

void CConvTempAnim::FixupTransDirectional(CTempTransform * pTmpTrans, const CInternKey1List &cRoll, const CTempTransform m_TmpTargetAnim[], const int32 num_targets)
{
	// search for target
	int i=0;
	bool bFoundTgt = false;
	while(i<num_targets && !bFoundTgt)
		if(pTmpTrans->m_s3dsName == m_TmpTargetAnim[i].m_s3dsName)
			bFoundTgt = true;
		else
			++i;

	// setup positions (and rotation)
	const CInternKey3List &kSrcKeyPos = pTmpTrans->m_cTranslation;
	CTempMat33List &kSrcKeyRot = pTmpTrans->m_cRotation;
	const CInternKey3List &kTgtKeyPos = (bFoundTgt) ? m_TmpTargetAnim[i].m_cTranslation : kSrcKeyPos;

	const int32 num_keys = (m_iFrameEnd - m_iFrameBegin) + 1;
	assert(kSrcKeyPos.GetNumKeys() == num_keys);
	assert(kSrcKeyRot.GetNumKeys() == num_keys);
	assert(kTgtKeyPos.GetNumKeys() == num_keys);
	assert(cRoll.GetNumKeys() == num_keys);

	// create new rotations
	for(int k=0; k<num_keys; k++)	
		kSrcKeyRot.m_pKeys[k] = CalcRot(kSrcKeyPos.m_pKeys[k], kTgtKeyPos.m_pKeys[k], cRoll.m_pKeys[k]);
}


const Mat33 CConvTempAnim::CalcRot(const Vec3 &vPos, const Vec3 &vTarget, const float roll)
{
	// calculate lookAt matrix
	Vec3 c1(1,0,0), c2(0,1,0), c3(0,0,1);
	Vec3 vDir = vTarget - vPos;
	if(vDir != Vec3(0,0,0))
	{
		vDir = Normalize(vDir);
		if(fabs(vDir*Vec3(0,0,1)) < 1)
			c1 = Normalize( Cross(vDir, Vec3(0,0,1)) );
		c2 = vDir;
		c3 = Normalize( Cross(c1, c2) );
	}

	// make result
	Mat33 m33;
	SetColumn(&m33, 0, c1);
	SetColumn(&m33, 1, c2);
	SetColumn(&m33, 2, c3);
	if(roll != 0)
	{
		// roll around nose
		Mat33 r33;
		const float fAngToRad = (float) (3.1415926535897932384626433832795 / 180);
		LoadRotationAxisAngle(&r33, GetColumn(m33, 1), roll * fAngToRad);
		m33 = r33 * m33;
	}
	
	// return result
	return m33;
}


// Clean up time
void CConvTempAnim::DownSampleMeshSplines(CTempMeshAnim * pMeshAnim)
{
	DownSampleTempTrans(pMeshAnim);
}

void CConvTempAnim::DownSampleCameraSplines(CTempCamAnim * pCamAnim)
{
	DownSampleTempTrans(pCamAnim);
	DownSampleKey1Lst(&pCamAnim->m_cFov);
	CloseDownKey1Lst(&pCamAnim->m_cRoll);
}

void CConvTempAnim::DownSampleOmniSplines(CTempOmniAnim * pOmniAnim)
{
	DownSampleTempTrans(pOmniAnim);
}

void CConvTempAnim::DownSampleSpotSplines(CTempSpotAnim * pSpotAnim)
{
	DownSampleTempTrans(pSpotAnim);
	DownSampleKey1Lst(&pSpotAnim->m_cHotSpot);
	DownSampleKey1Lst(&pSpotAnim->m_cFallOff);
	CloseDownKey1Lst(&pSpotAnim->m_cRoll);
}

void CConvTempAnim::DeleteCamTargetSplines(CTempCamTgtAnim * pCamTargetAnim)
{
	CloseDownTempTrans(pCamTargetAnim);
}
	
void CConvTempAnim::DeleteSpotTargetSplines(CTempSpotTgtAnim * pSpotTargetAnim)
{
	CloseDownTempTrans(pSpotTargetAnim);
}

void CConvTempAnim::DownSampleTempTrans(CTempTransform * pTransAnim)
{
	DownSampleKey1Lst(&pTransAnim->m_cUniScale);
	DownSampleKey3Lst(&pTransAnim->m_cTranslation);
	DownSampleTempMat33Lst(&pTransAnim->m_cRotation);
}

void CConvTempAnim::DownSampleKey1Lst(CInternKey1List * pKeyLst)
{
	// currently just using the simplest downsampling
	bool bFoundMisMatch = false;
	int i=1;
	while(i<pKeyLst->GetNumKeys() && !bFoundMisMatch)
		bFoundMisMatch = pKeyLst->GetKey(0)!=pKeyLst->GetKey(i++);

	// simple trick
	if(!bFoundMisMatch)
		pKeyLst->m_iNumKeys = 1;
}

void CConvTempAnim::DownSampleKey3Lst(CInternKey3List * pKeyLst)
{
	// currently just using the simplest downsampling
	bool bFoundMisMatch = false;
	int i=1;
	while(i<pKeyLst->GetNumKeys() && !bFoundMisMatch)
		bFoundMisMatch = pKeyLst->GetKey(0)!=pKeyLst->GetKey(i++);

	// simple trick
	if(!bFoundMisMatch)
		pKeyLst->m_iNumKeys = 1;
}

void CConvTempAnim::DownSampleTempMat33Lst(CTempMat33List * pKeyLst)
{
	// currently just using the simplest downsampling
	bool bFoundMisMatch = false;
	int i=1;
	const Vec3 c1 = GetColumn(pKeyLst->GetKey(0), 0);
	const Vec3 c2 = GetColumn(pKeyLst->GetKey(0), 1);
	const Vec3 c3 = GetColumn(pKeyLst->GetKey(0), 2);
	while(i<pKeyLst->GetNumKeys() && !bFoundMisMatch)
	{
		Mat33 m = pKeyLst->GetKey(i++);
		bFoundMisMatch = !((c1==GetColumn(m, 0)) && (c2==GetColumn(m, 1)) && (c3==GetColumn(m, 2)));
	}

	// simple trick
	if(!bFoundMisMatch)
		pKeyLst->m_iNumKeys = 1;
}

void CConvTempAnim::CloseDownTempTrans(CTempTransform * pTransAnim)
{
	CloseDownKey1Lst(&pTransAnim->m_cUniScale);
	CloseDownKey3Lst(&pTransAnim->m_cTranslation);
	CloseDownTempMat33Lst(&pTransAnim->m_cRotation);
}

void CConvTempAnim::CloseDownKey1Lst(CInternKey1List * pKeyLst)
{
	if(pKeyLst->m_pKeys != NULL) delete [] pKeyLst->m_pKeys;
	pKeyLst->m_iNumKeys = 0;
}

void CConvTempAnim::CloseDownKey3Lst(CInternKey3List * pKeyLst)
{
	if(pKeyLst->m_pKeys != NULL) delete [] pKeyLst->m_pKeys;
	pKeyLst->m_iNumKeys = 0;
}

void CConvTempAnim::CloseDownTempMat33Lst(CTempMat33List * pKeyLst)
{
	if(pKeyLst->m_pKeys != NULL) delete [] pKeyLst->m_pKeys;
	pKeyLst->m_iNumKeys = 0;
}

// user interface
const int32 CConvTempAnim::GetNumKey1s() const
{
	int32 res = -1;

	switch(attrib_type)
	{
		case FOV_GET:
			assert(node_mode==CAMERA_NODE);
			res = ((CTempCamAnim *) m_pTmpCurTrans)->m_cFov.GetNumKeys();
			break;
		case HOTSPOT_GET:
			assert(node_mode==SPOT_NODE);
			res = ((CTempSpotAnim *) m_pTmpCurTrans)->m_cHotSpot.GetNumKeys();
			break;
		case FALLOFF_GET:
			assert(node_mode==SPOT_NODE);
			res = ((CTempSpotAnim *) m_pTmpCurTrans)->m_cFallOff.GetNumKeys();
			break;
		case SCALE_GET:
			res = m_pTmpCurTrans->m_cUniScale.GetNumKeys();
			break;
		default:
			assert(false);
	}

	return res;
}

const int32 CConvTempAnim::GetNumKey3s() const
{
	assert(attrib_type==POSITION_GET);
	return m_pTmpCurTrans->m_cTranslation.GetNumKeys();
	
}

const int32 CConvTempAnim::GetNumMat33s() const
{
	assert(attrib_type==ROTATION_GET);
	return m_pTmpCurTrans->m_cRotation.GetNumKeys();
}

const float CConvTempAnim::GetKey1(const int32 index) const
{
	float res = -1;
	
	switch(attrib_type)
	{
		case FOV_GET:
			assert(node_mode==CAMERA_NODE);
			res = ((CTempCamAnim *) m_pTmpCurTrans)->m_cFov.GetKey(index);
			break;
		case HOTSPOT_GET:
			assert(node_mode==SPOT_NODE);
			res = ((CTempSpotAnim *) m_pTmpCurTrans)->m_cHotSpot.GetKey(index);
			break;
		case FALLOFF_GET:
			assert(node_mode==SPOT_NODE);
			res = ((CTempSpotAnim *) m_pTmpCurTrans)->m_cFallOff.GetKey(index);
			break;
		case SCALE_GET:
			res = m_pTmpCurTrans->m_cUniScale.GetKey(index);
			break;
		default:
			assert(false);
	}

	return res;
}

const Vec3 CConvTempAnim::GetKey3(const int32 index) const
{
	assert(attrib_type==POSITION_GET);
	return m_pTmpCurTrans->m_cTranslation.GetKey(index);
}

const Mat33 CConvTempAnim::GetMat33(const int32 index) const
{
	assert(attrib_type==ROTATION_GET);
	return m_pTmpCurTrans->m_cRotation.GetKey(index);
}


string CConvTempAnim::GetName() const
{
	switch(node_mode)
	{
		case MESH_NODE:
			return ((CTempMeshAnim *) m_pTmpCurTrans)->m_s3dsName;
		case CAMERA_NODE:
			return ((CTempCamAnim *) m_pTmpCurTrans)->m_s3dsName;
		case OMNI_NODE:
			return ((CTempOmniAnim *) m_pTmpCurTrans)->m_s3dsName;
		case SPOT_NODE:
			return ((CTempSpotAnim *) m_pTmpCurTrans)->m_s3dsName;
		default:
			assert(false);
			return string("arnes deller\n"); 
	}
}

string CConvTempAnim::GetTrackInstName() const
{
	switch(node_mode)
	{
		case MESH_NODE:
			return ((CTempMeshAnim *) m_pTmpCurTrans)->m_s3dsTrackInstName;
		case CAMERA_NODE:
			return ((CTempCamAnim *) m_pTmpCurTrans)->m_s3dsTrackInstName;
		case OMNI_NODE:
			return ((CTempOmniAnim *) m_pTmpCurTrans)->m_s3dsTrackInstName;
		case SPOT_NODE:
			return ((CTempSpotAnim *) m_pTmpCurTrans)->m_s3dsTrackInstName;
		default:
			assert(false);
			return string("arnes deller\n"); 
	}
}


// constructor
CConvTempAnim::CConvTempAnim()
{
	m_iNumMeshAnims = 0;
	m_iNumCamAnims = 0;
	m_iNumOmniAnims = 0;
	m_iNumSpotAnims = 0;
	m_iNumCamTargetAnims = 0;
	m_iNumSpotTargetAnims = 0;
	
	m_pTmpMeshAnim = NULL;
	m_pTmpCamAnim = NULL;
	m_pTmpOmniAnim = NULL;
	m_pTmpSpotAnim = NULL;
	m_pTmpCamTargetAnim = NULL;
	m_pTmpSpotTargetAnim = NULL;

	m_pTmpCurTrans = NULL;
	node_mode = NODE_UNKNOWN;
	attrib_type = ATTRIB_UNKNOWN;
}
