#include "inputcontainer.h"
#include <stdio.h>
#include <memory.h>
#include <assert.h>

// Public API
const int32 CInputContainer::GetNum3dsInputFiles()
{
	return m_iInListLen;
}

const char * CInputContainer::Get3dsInputPath(const int32 i3dsInNumber)
{
	assert(i3dsInNumber<GetNum3dsInputFiles());
	return m_pInLst[i3dsInNumber];
}

const char * CInputContainer::GetOutputFolder()
{
	return m_pOutPath;
}

const int32 CInputContainer::GetMaxVertsInStrip()
{
	return (m_iStripMaxVerts < 3) ? -1 : m_iStripMaxVerts;
}



// Secret API
bool CInputContainer::AddInputFile(const char * cptr, const int32 length)
{
	return AddStringToSpecList(cptr, length, &m_pInLst, &m_iInListLen);
}

bool CInputContainer::AddTextureFolder(const char * cptr, const int32 length)
{
	return CTexturePropContainer::AddTextureFolder(cptr, length);
}

bool CInputContainer::SetStripMaxVerts(const int32 strip_maxverts)
{
	m_iStripMaxVerts = strip_maxverts;
	return true;
}


bool CInputContainer::SetOutputFolder(const char * cptr, const int32 length)
{
	// check if output has been set already
	if(m_pOutPath!=NULL)
	{
		delete [] m_pOutPath;
		m_pOutPath = NULL;
	}

	// fetch string
	bool bRes = false;

	if((m_pOutPath = new char[length+1]) != NULL)
	{
		memcpy(m_pOutPath, cptr, length);
		m_pOutPath[length] = '\0';
		bRes = true;
	}

	return bRes;
}

bool CInputContainer::SetDefaultTextureProps(const CTextureProps &texProps)
{
	m_cDefaultProps = texProps;
	return true;
}

bool CInputContainer::AddSpecificTextureProps(const CTextureProps &texProps)
{
	const char * cptr = texProps.GetTexPath();

	CTextureProps * pLocTexProp = PriTexPropLocate( texProps.GetTexPath() );
	bool bSuccess=false;

	// insert props
	if(pLocTexProp!=NULL)
	{	
		*pLocTexProp = texProps;
		bSuccess=true;
		printf("Warning! Texture props have been set before on: %s\n", texProps.GetTexPath());
	}
	else
	{
		CTextureProps * pOldTexProps = m_pPropsOfSpecTexts;
		m_pPropsOfSpecTexts = NULL;
		if((m_pPropsOfSpecTexts = new CTextureProps[m_iNumPropSpecTexts+1]) != NULL)
		{
			for(int i=0; i<m_iNumPropSpecTexts; i++)
				m_pPropsOfSpecTexts[i] = pOldTexProps[i];
			if(pOldTexProps != NULL) delete [] pOldTexProps;
			m_pPropsOfSpecTexts[m_iNumPropSpecTexts++] = texProps;
			
			bSuccess = true;
		}
		else
		{
			// cleanup, we failed
			m_pPropsOfSpecTexts = pOldTexProps;
		}
	}

	return bSuccess;
}


// constructor and destructor
CInputContainer::CInputContainer()
{
	m_iInListLen = 0;
	m_pInLst = NULL;

	m_iStripMaxVerts = -1;
	
	// output folder
	m_pOutPath = NULL;
}

CInputContainer::~CInputContainer()
{


}
