#ifndef __KEYF_H__
#define __KEYF_H__

#include "VECTOR.h"

struct controller
{
	controller(const int kn, const float t, const float c, const float b, const float et, const float ef)
	{
		key_num = kn;
		tens = t;
		cont = c;
		bias = b;
		easeto = et;
		easefrom = ef;
	}
	
	int key_num;
	float tens, cont, bias, easeto, easefrom;
};

struct keyf1 : controller
{
	keyf1(const int kn, const float t, const float c, const float b, const float et, const float ef, const float X)
	: controller(kn, t, c, b, et, ef), x(X) {}
	
	float x;
};

struct keyf3 : controller
{
	keyf3(const int kn, const float t, const float c, const float b, const float et, const float ef, const float X, const float Y, const float Z)
	: controller(kn, t, c, b, et, ef), x(X), y(Y), z(Z) {}

	float x, y, z;
};

struct keyf4 : controller
{
	keyf4(const int kn, const float t, const float c, const float b, const float et, const float ef, const float X, const float Y, const float Z, const float W)
	: controller(kn, t, c, b, et, ef), x(X), y(Y), z(Z), w(W) {}
	
	float x, y, z, w;
};

class C3DTrackInfo
{
	friend class C3DModel;
	friend class CConvTempAnim;
	
public:
	const int GetNumPosKeys() const { return pos.size(); }
	const int GetNumRotKeys() const { return rot.size(); }
	const int GetNumScaleKeys() const { return scale.size(); }
	const keyf3 GetPosKey(const int index) const { return pos[index]; }
	const keyf4 GetRotKey(const int index) const { return rot[index]; }
	const keyf3 GetScaleKey(const int index) const { return scale[index]; }

	const VECTOR GetPivotPoint() const { return pivot_point; }
	const int GetHierarchyNum() const { return hierarchy_num; }
	const string GetNameID() const { return m_sIDName; }
	const string GetTrackInstName() const { return m_sTrackInstName; }
	const int GetNodeID() const { return node_id; }

	C3DTrackInfo() : pivot_point(0,0,0), node_id(0), hierarchy_num(-1) {}
	
protected:
	vector <keyf3> pos;
	vector <keyf4> rot;
	vector <keyf3> scale;
	VECTOR pivot_point;	

	int node_id;				// 0xb030
	int hierarchy_num;		// strange system to ID location in hierachy.
	string m_sIDName;
	string m_sTrackInstName;	// track instance name
};

class C3DTrackCamera : public C3DTrackInfo
{
	friend class C3DModel;
	friend class CConvTempAnim;
	
public:
	const int GetNumRollKeys() const { return roll.size(); }
	const int GetNumFovKeys() const { return fov.size(); }
	const keyf1 GetRollKey(const int index) const { return roll[index]; }
	const keyf1 GetFovKey(const int index) const { return fov[index]; }

protected:
	vector <keyf1> roll;
	vector <keyf1> fov;
};

class C3DTrackOmni : public C3DTrackInfo
{
	friend class C3DModel;
	
protected:
};

class C3DTrackSpot : public C3DTrackInfo
{
	friend class C3DModel;

public:
	const int GetNumRollKeys() const { return roll.size(); }
	const int GetNumHotSpotKeys() const { return hotspot.size(); }
	const int GetNumFallOffKeys() const { return falloff.size(); }
	const keyf1 GetRollKey(const int index) const { return roll[index]; }
	const keyf1 GetHotSpotKey(const int index) const { return hotspot[index]; }
	const keyf1 GetFallOffKey(const int index) const { return falloff[index]; }
	
protected:
	vector <keyf1> roll;
	vector <keyf1> hotspot;
	vector <keyf1> falloff;
};

class C3DTrackCamTarget : public C3DTrackInfo
{
	friend class C3DModel;
	
protected:
};

class C3DTrackSpotTarget : public C3DTrackInfo
{
	friend class C3DModel;
	
protected:
};


#endif
